/*

universal primitive 0x0a        - done
link condition                  - done
extra null's  TransitionTo
ActionSlots
OrigContent :ContentRef
tabs/newlines/brackets in tokengroups
extra defaultattributes         - done


*/

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <snmp.h>

#include "asn_decode.h"


#define APPLICATION     0
#define SCENE           1


/* in <snmp.h>
#define ASN_UNIVERSAL                   0x00
#define ASN_APPLICATION                 0x40
#define ASN_CONTEXT                     0x80
#define ASN_PRIVATE                     0xC0

#define ASN_PRIMITIVE                   0x00
#define ASN_CONSTRUCTOR                 0x20
*/

#define ASN_BOOLEAN                     0x01
#define ASN_BITSTRING                   0x03
#define ASN_ENUMERATED                  0x0a

/*

class of tag
    universal   - 00
        1 - boolean
        2 - integer
        3 - bitstring
        4 - octetstring
        5 - null
        6 - object identifier
        7 - object descriptor
        8 - external
        9 - real
        10 - enumerated
        11 -
        12-15 - reserved
        16 - sequence
        17 - set
        18-22 - character string
        23-24 - time
        25-27 - character string
        28-   - reserved

    application - 40
    private     - 80
    context     - c0


encoding
    identifier octets
        bit 8,7   => class
        bit 6     => p/c,  0 => primitive encoding, 1=> constructed encoding
        bit 5-0   => tag
        for tags >= 31, add leading octet
            = <class of tag (00,01,10,11)> <(0,1)> <11111>
        subsequent octets
            bit 8 = 1, except for last identifier octet
            bit 7-1 - tag number (1st NOT 00)

        e.g.
            low tag:  [<class><p/c><tag>]
            high tag: [<class><p/c><11111>] [1xxxxxxx] [1xxxxxxx] [0xxxxxxx]

    length octets
        2 forms, definate, indefinate
        use definate, if primitive encoding
        use either, if constructed and all available
        use indefinate, if constucted and not all available
        definate
            short form, length = 0-127,
                bit8    => 0
                bit 7-1 => length
            long form, initial octet + subsequent octets
                initial octet, (0xff not used)
                    bit 8   => 1
                    bit 7-1 => number of subsequent octets
                subsequent octets
                    bit 8-1 => length
        indefinate
            single octet = 0x80
            ... content octets ...
            end of contents octets = 0x00, 0x00    (equiv. to primitive, tag 0, value 0)

    content octets
        data

    end of contents octets
        end of contents octets = 0x00, 0x00    (equiv. to primitive, tag 0, value 0)



*/





#define DESCLEN 40



// EnumType
typedef struct
{
    UCHAR id;
    UCHAR description[DESCLEN];
} EnumType;


// EventType
EnumType EventType[] =
{
// types
    { 0x1f, "UNDEFINED"                },
    { 0x01, "IsAvailable"              },
    { 0x02, "ContentAvailable"         },
    { 0x03, "IsDeleted"                },
    { 0x04, "IsRunning"                },
    { 0x05, "IsStopped"                },
    { 0x06, "UserInput"                },
    { 0x07, "AnchorFired"              },
    { 0x08, "TimerFired"               },
    { 0x09, "AsynchStopped"            },
    { 0x0a, "InteractionCompleted"     },
    { 0x0b, "TokenMovedFrom"           },
    { 0x0c, "TokenMovedTo"             },
    { 0x0d, "StreamEvent"              },
    { 0x0e, "StreamPlaying"            },
    { 0x0f, "StreamStopped"            },
    { 0x10, "CounterTrigger"           },
    { 0x11, "HighlightOn"              },
    { 0x12, "HighlightOff"             },
    { 0x13, "CursorEnter"              },
    { 0x14, "CursorLeave"              },
    { 0x15, "IsSelected"               },
    { 0x16, "IsDeselected"             },
    { 0x17, "TestEvent"                },
    { 0x18, "FirstItemPresented"       },
    { 0x19, "LastItemPresented"        },
    { 0x1a, "HeadItems"                },
    { 0x1b, "TailItems"                },
    { 0x1c, "ItemSelected"             },
    { 0x1d, "ItemDeselected"           },
    { 0x1e, "EntryFieldFull"           },
    { 0x1f, "EngineEvent"              }
};



// Justification
EnumType Justification[] =
{
// types
    { 0x04, "UNDEFINED"                },
    { 0x01, "start"                    },
    { 0x02, "end"                      },
    { 0x03, "centre"                   },
    { 0x04, "justified"                }
};


// LineOrientation
EnumType LineOrientation[] =
{
// types
    { 0x02, "UNDEFINED"                },
    { 0x01, "vertical"                 },
    { 0x02, "horizontal"               }
};


// StartCorner
EnumType StartCorner[] =
{
// types
    { 0x04, "UNDEFINED"                },
    { 0x01, "UpperLeft"                },
    { 0x02, "UpperRight"               },
    { 0x03, "LowerLeft"                },
    { 0x04, "LowerRight"               }
};


// Storage
EnumType Storage[] =
{
// types
    { 0x02, "UNDEFINED"                },
    { 0x01, "memory"                   },
    { 0x02, "stream"                   }
};



// Termination
EnumType Termination[] =
{
// types
    { 0x02, "UNDEFINED"                },
    { 0x01, "freeze"                   },
    { 0x02, "disappear"                }
};


// Orientation
EnumType Orientation[] =
{
// types
    { 0x04, "UNDEFINED"                },
    { 0x01, "left"                     },
    { 0x02, "right"                    },
    { 0x03, "up"                       },
    { 0x04, "down"                     }
};


// SliderStyle
EnumType SliderStyle[] =
{
// types
    { 0x03, "UNDEFINED"                },
    { 0x01, "normal"                   },
    { 0x02, "thermometer"              },
    { 0x03, "proportional"             }
};


// ButtonStyle
EnumType ButtonStyle[] =
{
// types
    { 0x03, "UNDEFINED"                },
    { 0x01, "pushbutton"               },
    { 0x02, "radiobutton"              },
    { 0x03, "checkbox"                 }
};


// InputType
EnumType InputType[] =
{
// types
    { 0x04, "UNDEFINED"                },
    { 0x01, "alpha"                    },
    { 0x02, "numeric"                  },
    { 0x03, "any"                      },
    { 0x04, "listed"                   }
};




// Class
EnumType Class[] =
{
    { 0xe8, "UNDEFINED"                },
    { 0x00, "Application"              },
    { 0x01, "Scene"                    },
    { 0x02, "StdID"                    },
    { 0x03, "StandardVersion"          },
    { 0x04, "ObjectInformation"        },
//    { 0x04, "ObjectInfo"               },
    { 0x05, "OnStartup"                },
    { 0x06, "OnCloseDown"              },
    { 0x07, "OriginalGroupCachePriority" },
//    { 0x07, "OrigGCPriority"           },
    { 0x08, "Items"                    },
    { 0x09, "ResidentPrg"              },
    { 0x0a, "RemotePrg"                },
    { 0x0b, "InterchangedPrg"          },
    { 0x0c, "Palette"                  },
    { 0x0d, "Font"                     },
    { 0x0e, "CursorShape"              },
    { 0x0f, "BooleanVar"               },
    { 0x10, "IntegerVar"               },
    { 0x11, "OStringVar"               },
    { 0x12, "ObjectRefVar"             },
    { 0x13, "ContentRefVar"            },
    { 0x14, "Link"                     },
    { 0x15, "Stream"                   },
    { 0x16, "Bitmap"                   },
    { 0x17, "LineArt"                  },
    { 0x18, "DynamicLineArt"           },
    { 0x19, "Rectangle"                },
    { 0x1a, "Hotspot"                  },
    { 0x1b, "SwitchButton"             },
    { 0x1c, "PushButton"               },
    { 0x1d, "Text"                     },
    { 0x1e, "EntryField"               },
    { 0x1f, "HyperText"                },
    { 0x20, "Slider"                   },
    { 0x21, "TokenGroup"               },
    { 0x22, "ListGroup"                },
    { 0x23, "OnSpawnCloseDown"         },
    { 0x24, "OnRestart"                },
    { 0x25, "DefaultAttributes"        },
    { 0x26, "CharacterSet"             },
    { 0x27, "BackgroundColour"         },
    { 0x28, "TextCHook"                },
    { 0x29, "TextColour"               },
    { 0x2a, "Font"                     },
    { 0x2b, "FontAttributes"           },
    { 0x2c, "InterchangedProgramContentHook" },
    { 0x2d, "StreamCHook"              },
    { 0x2e, "BitmapCHook"              },
    { 0x2f, "LineArtCHook"             },
    { 0x30, "ButtonRefColour"          },
    { 0x31, "HighlightRefColour"       },
    { 0x32, "SliderRefColour"          },
    { 0x33, "InputEventReg"            },
    { 0x34, "SceneCS"                  },
    { 0x35, "AspectRation"             },
    { 0x36, "MovingCursor"             },
    { 0x37, "NextScenes"               },
    { 0x38, "InitiallyActive"          },
    { 0x39, "ContentHook"              },
//    { 0x39, "CHook"                    },
    { 0x3a, "OrigContent"              },
    { 0x3b, "Shared"                   },
    { 0x3c, "ContentSize"              },
    { 0x3d, "ContentCachePriority"     },
//    { 0x3d, "CCPriority"               },
    { 0x3e, "LinkCondition"            },
    { 0x3f, "LinkEffect"               },
    { 0x40, "Name"                     },
    { 0x41, "InitiallyAvailable"       },
    { 0x42, "ProgramConnectionTag"     },
    { 0x43, "OrigValue"                },
    { 0x44, "ObjectRef"                },
    { 0x45, "ContentRef"               },
    { 0x46, "MovementTable"            },
    { 0x47, "TokenGroupItems"          },
    { 0x48, "NoTokenActionSlots"       },
    { 0x49, "Positions"                },
    { 0x4a, "WrapAround"               },
    { 0x4b, "MultipleSelection"        },
    { 0x4c, "OrigBoxSize"              },
    { 0x4d, "OrigPosition"             },
    { 0x4e, "OriginalPaletteRef"       },
    { 0x4f, "Tiling"                   },
    { 0x50, "OriginalTransparency"     },
    { 0x51, "BBBox"                    },
    { 0x52, "OrigLineWidth"            },
    { 0x54, "OrigRefLineColour"        },
    { 0x55, "OrigRefFillColour"        },
    { 0x56, "OriginalFont"             },
    { 0x57, "HJustification"           },
    { 0x58, "VJustification"           },
    { 0x59, "LineOrientation"          },
    { 0x5a, "StartCorner"              },
    { 0x5b, "TextWrapping"             },
    { 0x5c, "Multiplex"                },
    { 0x5d, "Storage"                  },
    { 0x5e, "Looping"                  },
    { 0x5f, "Audio"                    },
    { 0x60, "Video"                    },
    { 0x61, "RTGraphics"               },
    { 0x62, "ComponentTag"             },
    { 0x63, "OriginalVolume"           },
    { 0x64, "Termination"              },
    { 0x65, "EngineResp"               },
    { 0x66, "orientation"              },
    { 0x67, "MaxValue"                 },
    { 0x68, "MinValue"                 },
    { 0x69, "InitialValue"             },
    { 0x6a, "InitialPortion"           },
    { 0x6b, "StepSize"                 },
    { 0x6c, "SliderStyle"              },
    { 0x6d, "InputType"                },
    { 0x6e, "CharList"                 },
    { 0x6f, "ObscuredInput"            },
    { 0x70, "MaxLength"                },
    { 0x71, "OriginalLabel"            },
    { 0x72, "ButtonStyle"              },
    { 0x73, "Activate"                 },
    { 0x74, "Add"                      },
    { 0x75, "AddItem"                  },
    { 0x76, "Append"                   },
    { 0x77, "BringToFront"             },
    { 0x78, "Call"                     },
    { 0x79, "CallActionSlot"           },
    { 0x7a, "Clear"                    },
    { 0x7b, "Clone"                    },
    { 0x7c, "CloseConnection"          },
    { 0x7d, "Deactivate"               },
    { 0x7e, "DelItem"                  },
    { 0x7f, "Deselect"                 },
    { 0x80, "DeselectItem"             },
    { 0x81, "Divide"                   },
    { 0x82, "DrawArc"                  },
    { 0x83, "DrawLine"                 },
    { 0x84, "DrawOval"                 },
    { 0x85, "DrawPolygon"              },
    { 0x86, "DrawPolyLine"             },
    { 0x87, "DrawRectangle"            },
    { 0x88, "DrawSector"               },
    { 0x89, "Fork"                     },
    { 0x8a, "GetAvailabilityStatus"    },
    { 0x8b, "GetBoxSize"               },
    { 0x8c, "GetCellItem"              },
    { 0x8d, "GetCursorPosition"        },
    { 0x8e, "GetEngineSupport"         },
    { 0x8f, "GetEntryPoint"            },
    { 0x90, "GetFillColour"            },
    { 0x91, "GetFirstitem"             },
    { 0x92, "GetHighlightStatus"       },
    { 0x93, "GetInteractionStatus"     },
    { 0x94, "GetItemStatus"            },
    { 0x95, "GetLabel"                 },
    { 0x96, "GetLastAnchorFired"       },
    { 0x97, "GetLineColour"            },
    { 0x98, "GetLineStyle"             },
    { 0x99, "GetLineWidth"             },
    { 0x9a, "GetLineItem"              },
    { 0x9b, "GetLineSize"              },
    { 0x9c, "GetOverwriteMode"         },
    { 0x9d, "GetPortion"               },
    { 0x9e, "GetPosition"              },
    { 0x9f, "GetRunningStatus"         },
    { 0xa0, "GetSelectionStatus"       },
    { 0xa1, "GetSlidervalue"           },
    { 0xa2, "GetTextContent"           },
    { 0xa3, "GetTextData"              },
    { 0xa4, "GetTokenPosition"         },
    { 0xa5, "GetVolume"                },
    { 0xa6, "Launch"                   },
    { 0xa7, "LockScreen"               },
    { 0xa8, "Modulo"                   },
    { 0xa9, "Move"                     },
    { 0xaa, "MoveTo"                   },
    { 0xab, "Multiply"                 },
    { 0xac, "OpenConnection"           },
    { 0xad, "Preload"                  },
    { 0xae, "PutBefore"                },
    { 0xaf, "PutBehind"                },
    { 0xb0, "Quit"                     },
    { 0xb1, "ReadPersistent"           },
    { 0xb2, "Run"                      },
    { 0xb3, "ScaleBitmap"              },
    { 0xb4, "ScaleVideo"               },
    { 0xb5, "ScrollItems"              },
    { 0xb6, "Select"                   },
    { 0xb7, "SelectItem"               },
    { 0xb8, "SendEvent"                },
    { 0xb9, "SendToBack"               },
    { 0xba, "SetBoxSize"               },
    { 0xbb, "SetCachePriority"         },
    { 0xbc, "SetCounterEndPosition"    },
    { 0xbd, "SetCounterPosition"       },
    { 0xbe, "SetCounterTrigger"        },
    { 0xbf, "SetCursorposition"        },
    { 0xc0, "SetCursorShape"           },
    { 0xc1, "SetData"                  },
    { 0xc2, "SetEntryPoint"            },
    { 0xc3, "SetFillColour"            },
    { 0xc4, "SetFirstIem"              },
    { 0xc5, "SetFontref"               },
    { 0xc6, "SetHighlightStatus"       },
    { 0xc7, "SetInteractionStatus"     },
    { 0xc8, "SetLabel"                 },
    { 0xc9, "SetLineColour"            },
    { 0xca, "SetLineStyle"             },
    { 0xcb, "SetLineWidth"             },
    { 0xcc, "SetOverwriteMode"         },
    { 0xcd, "SetPaletteRef"            },
    { 0xce, "SetPortion"               },
    { 0xcf, "SetPosition"              },
    { 0xd0, "SetSlidervalue"           },
    { 0xd1, "SetSpeed"                 },
    { 0xd2, "SetTimer"                 },
    { 0xd3, "SetTransparency"          },
    { 0xd4, "SetVariable"              },
    { 0xd5, "SetVolume"                },
    { 0xd6, "Spawn"                    },
    { 0xd7, "Step"                     },
    { 0xd8, "Stop"                     },
    { 0xd9, "StorePersistent"          },
    { 0xda, "Subtract"                 },
    { 0xdb, "TestVariable"             },
    { 0xdc, "Toggle"                   },
    { 0xdd, "ToggleItem"               },
    { 0xde, "TransitionTo"             },
    { 0xdf, "Unload"                   },
    { 0xe0, "UnlockScreen"             },

    { 0xe1, "GBoolean"                 },
    { 0xe2, "GInteger"                 },
    { 0xe3, "GOctetString"             },
    { 0xe4, "GObjectRef"               },
    { 0xe5, "GContentRef"              },

    { 0xe7, "NewAbsoluteColour"        },

    { 0xea, "NewRefContent"            },
    { 0xeb, "NewCCPriority"            },
    { 0xec, "IndirectRef"              }


};


char * tab(void);
char * getClass(FILE *Op, int n);
int asn_boolean(FILE *Op, UCHAR *buf);
int asn_integer(FILE *Op, UCHAR *buf, int length);
int asn_bitstring(FILE *Op, UCHAR *buf, int length);
int asn_octetstring(FILE *Op, UCHAR *buf, int length);
int asn_octetstring_adjust(FILE *Op, UCHAR *buf, int length);
int asn_null(FILE *Op, UCHAR *buf);
int asn_objectidentifier(FILE *Op, UCHAR *buf, int length);
int mheg_enumtype(FILE *Op, EnumType *qenum, UCHAR *buf, int length);
int get_default_attributes(UCHAR *buf, int length1);


int tabl = 0;
char tabs[50];

int last_tag = 0;
int enable_actionslots = 0;
int maja = 1;
int asn_debug = 0;


int default_font_attribute_length = 0;
UCHAR default_font_attribute[40];
int font_attribute_found;

int default_text_colour_length = 0;
UCHAR default_text_colour[40];
int text_colour_found;
/*
UCHAR buf[0x10000];
int   blen;
int   offset;
extern int asn_debug = 0;
int main(int argc, char *argv[])
{
    FILE *Ip;
    FILE *Op;
    UCHAR  data1;
    USHORT data2;
    UCHAR exp;
    int err;
    int length;
    int i,j;
    char infile[100];
    char outfile[100];
    int files = 0;

// process command line
    strcpy(infile,"welcome.asn");
    strcpy(outfile,"welcome.mh5");
    for(i=1; i<argc; i++)
    {
        j = 0;
        if ((argv[i][0] == '-') || (argv[i][0] == '/'))
        {
            j++;
            // skip spaces ...
            while(argv[i][j] == '\0') {i++;j=0;}
            switch (argv[i][j++])
            {
                case 'v' :
                    asn_debug = 0xff;
                    break;
                case 'm' :
                    maja = 0;
                    break;
                case 'h' :
                case '?' :
                default  :
//                    usage();
                    return 1;
            }
        }
        else
        {
            // set ip filename from command line parameter
            if (files == 0) sscanf(argv[i],"%s",infile);
            if (files == 1) sscanf(argv[i],"%s",outfile);
            files++;
        }
    }



    Ip = fopen(infile,"rb");
    if (!Ip)
    {
        printf("Unable to open file : %s\n",infile);
        return -1;
    }
    blen = fread(buf,1,0x10000,Ip);
    fclose(Ip);

    if ((buf[0] != 0xa0) && (buf[0] != 0xa1))
    {
        printf("asn: %s - not MHEG-5 ASN.1 format\n", infile);
        return -1;
    }
    Op = fopen(outfile,"wt");
    if (!Op)
    {
        printf("Unable to open file : %s\n",outfile);
        return -1;
    }
//printf("outfile=%s - %x\n",outfile,Op);


    offset = 0;
    i = offset;
    j = 0;

    exp = 0xa1;
    data1 = buf[offset];
    if ((data1 != 0xa0) && (data1 != 0xa1))
    {
        printf("Offset %x, Read %x, Expected %x\n",offset,data1,exp);
        return -1;
    }
    j += asn_decode(Op,&buf[i+j]);


    fprintf(Op,"\n");
    if (Op) fclose(Op);


    return 0;
}

*/










/*

InterchangedObject[] =
{
    { 0x00, ApplicationClass },
    { 0x01, SceneClass },
};

GroupClass[] =
{

    StdIdentifer            StdIdentifier   OPTIONAL
    StdVersion              integer         OPTIONAL
    ObjectInfo              octet_string    OPTIONAL
    OnStartUp               ActionClass     OPTIONAL
    OnCloseDown             ActionClass     OPTIONAL
    OriginalGCPriority      integer [127]
    Items                   seq. of GroupItems


};

ApplicationClass[] =
{
    GroupClass
    OnSpawnCloseDown,   ActionClass, OPTIONAL,
    OnRestart,          ActionClass, OPTIONAL,
    default attributes
    { 0x00, ApplicationClass },
    { 0x01, SceneClass },
};

*/







char * tab(void)
{
    int i;

    for (i=0; i<(tabl*4); i++)
    {
        tabs[i] = ' ';
    }
    tabs[i] = '\0';
    return tabs;
}





char * getClass(FILE *Op, int n)
{
    int i;

    for(i=1; i<(sizeof(Class) / (DESCLEN+1)); i++)
    {
        if (Class[i].id == n)
            return Class[i].description;
    }
    fprintf(Op," %02x ",n);
    return Class[0].description;
}


int asn_boolean(FILE *Op, UCHAR *buf)
{
//    printf("ASN_BOOLEAN     ");
    if (asn_debug & 1) printf("[%02x] ",buf[0]);
    fprintf(Op,"%s ",(buf[0] == 0) ? "false" : "true");
    return 1;
}

int asn_integer(FILE *Op, UCHAR *buf, int length)
{
    int val,j;

//    printf("ASN_INTEGER     ");
    val = 0;
    for(j=0; j<length; j++)
    {
        if (asn_debug & 1) printf("[%02x] ",buf[j]);
    }
    for(j=0; j<length; j++)
    {
        val <<= 8;
        val |= buf[j];
    }
    // sign
    if (val & (0x80 << ((length-1)*8)) ) val = val-(length*0x100);
    fprintf(Op,"%d ",val);
    return length;
}

int asn_bitstring(FILE *Op, UCHAR *buf, int length)
{
    if (asn_debug & 1) printf("[%02x] ",buf[0]);
    printf("ASN_BITSTRING   ");
    return length;
}

int asn_octetstring(FILE *Op, UCHAR *buf, int length)
{
    int j;

    if (asn_debug & 1) printf("[%02x] ",buf[0]);
//    printf("ASN_OCTETSTRING ");
    fprintf(Op,"'");
    for(j=0; j<length; j++)
    {
        if (asn_debug & 1) printf("[%02x] ",buf[j]);
    }
    for(j=0; j<length; j++)
    {
        if ((buf[j] >= 0x20) && (buf[j] < 0x7f) && (buf[j] != 39))
            fprintf(Op,"%c",buf[j]);
        else
            fprintf(Op,"=%02X",buf[j]);
    }
    fprintf(Op,"' ");
    return length;
}

//FontAttributes
//    :FontAttributes 'plain.20.23.0'
//    :FontAttributes =00=18=18=00=00
// reduce text size by  20%

int asn_octetstring_adjust(FILE *Op, UCHAR *buf, int length)
{
    int j;
    int element = 0;
    int value;
    int numin;
    int scan = 0;

    if (asn_debug & 1) printf("[%02x] ",buf[0]);
//    printf("ASN_OCTETSTRING ");
    fprintf(Op,"'");
//printf("\nasn_octetstring_adjust(), length=%d ",length);
    for(j=0; j<length; j++)
    {
        if (asn_debug & 1) printf("[%02x] ",buf[j]);
//        printf("[%02x] ",buf[j]);
    }

    // crude, but if length > 5 assume 'plain.24.24.0' format
    if (length > 5)
    {
        for(j=0; j<length; j++)
        {
            if (buf[j] == '.')
            {
                scan = 0;
                if (element == 0)
                {
                    numin = sscanf(&buf[j],".%d",&value);
                    value = value*4/5;
                    fprintf(Op,".%d",value);
                    scan = 1;
                    element++;
                }
                else if (element == 1)
                {
                    numin = sscanf(&buf[j],".%d",&value);
                    value = value*4/5;
                    fprintf(Op,".%d",value);
                    scan = 1;
                    element++;
                }
                else
                {
                    if ((buf[j] >= 0x20) && (buf[j] < 0x7f) && (buf[j] != 39))
                        fprintf(Op,"%c",buf[j]);
                    else
                        fprintf(Op,"=%02X",buf[j]);
                }
            }
            else if (!scan)
            {
                if ((buf[j] >= 0x20) && (buf[j] < 0x7f) && (buf[j] != 39))
                    fprintf(Op,"%c",buf[j]);
                else
                    fprintf(Op,"=%02X",buf[j]);
            }
        }
    }
    // crude, but if length <=5  assume '=00=18=18=00=00' format
    else
    {
        for(j=0; j<length; j++)
        {
            value = buf[j];
            if (j==1)
            {
                value = value*4/5;
            }
            else if (j==2)
            {
                value = value*4/5;
            }
//printf("\navalue=%x ",value);
            if ((value >= 0x20) && (value < 0x7f) && (value != 39))
                fprintf(Op,"%c",value);
            else
                fprintf(Op,"=%02X",value);
        }
    }

    fprintf(Op,"' ");

    return length;
}

int asn_null(FILE *Op, UCHAR *buf)
{
    if (asn_debug & 1) printf("[%02x] ",buf[0]);
//    printf("ASN_NULL        ");
    fprintf(Op,"null ");
    return 0;
}

int asn_objectidentifier(FILE *Op, UCHAR *buf, int length)
{
    if (asn_debug & 1) printf("[%02x] ",buf[0]);
    printf("ASN_OBJECTIDENTIFIER ");
    return length;
}


int mheg_enumtype(FILE *Op, EnumType *qenum, UCHAR *buf, int length)
{
    int i;


    if (asn_debug & 1) printf("[%02x] ",buf[0]);

    for(i=1; i<=qenum[0].id; i++)
    {
        if (qenum[i].id == buf[0])
        {
            fprintf(Op,"%s ",qenum[i].description);
            return 1;
        }
    }
    fprintf(Op," %02x ",buf[0]);
    printf("XXX %s ",qenum[0].description);

    return 1;
}



int asn_decode(FILE *Op, UCHAR *buf)
{
    int state = 0;
    int class;
    int encoding;
    int tag;
    int length;
    int subsequent = 0;
    int content = 0;
    int i,j;

//fprintf(Op,"[asn_decode()]");
//    printf("\n%s ",tab());
//    printf("asn_decode() ");
//    tabl ++;


    i = 0;

    // identifier octet
//    printf("\n%s ",tab());
    if (asn_debug & 1) printf("[%02x] ",buf[i]);
    class = buf[i] & 0xc0;
    encoding = buf[i] & 0x20;
    tag = buf[i] & 0x1f;
//    printf("[[buf[%d]:%02x, class:%02x, encoding:%02x, tag:%02x]]",i,buf[i],class,encoding,tag);
    i++;


    // subsequent identifier octets
    if (tag == 0x1f)
    {
        tag = 0;
        do
        {
            if (asn_debug & 1) printf("[%02x] ",buf[i]);
            tag = (tag << 7) + (buf[i] & 0x7f);
            subsequent = buf[i] & 0x80;
            i++;
        } while (subsequent);
    }
//    printf("[buf[%d]:%02x, class:%02x, encoding:%02x, tag:%02x]",i,buf[i],class,encoding,tag);


    // length
//    printf("\n%s ",tab());
    if (asn_debug & 1) printf("[%02x] ",buf[i]);
    length = buf[i] & 0x7f;
    subsequent = buf[i] & 0x80;
    i++;

    // subsequent length
    if (subsequent)
    {
        subsequent = length;
        length = 0;
        do
        {
            if (asn_debug & 1) printf("[%02x] ",buf[i]);
            length = (length << 8) + (buf[i] & 0xff);
            subsequent --;
            i++;
        }
        while (subsequent);
    }
//    printf("length:%02x",length);
//    printf("\n%s",tab());


    // content
//fprintf(Op,"[class:%02x, encoding:%02x, tag:%02x]",class,encoding,tag);

//fprintf(Op,"[tag=%x]",tag);
//    while(length)
    do
    {
        switch(class)
        {
            case 0x00:  // universal
                switch(encoding)
                {
                    case ASN_PRIMITIVE:
                        switch(tag)
                        {
                            case 0x00:  // end of contents
                                if (asn_debug & 1) printf("[%02x] ",buf[i]);
//                                i ++;
                                length = 0;
                                return 0;
                                break;

                            case ASN_BOOLEAN:           //0x01
                                i += asn_boolean(Op,&buf[i]);
                                length = 0;
                                break;

                            case ASN_INTEGER:           //0x02
                                i += asn_integer(Op,&buf[i],length);
                                length = 0;
                                break;

                            case ASN_BITSTRING:         //0x03
                                i += asn_bitstring(Op,&buf[i],length);
                                length = 0;
                                break;

                            case ASN_OCTETSTRING:       //0x04
                                i += asn_octetstring(Op,&buf[i],length);
                                length = 0;
                                break;

                            case ASN_NULL:              //0x05
                                i += asn_null(Op,&buf[i]);
                                length = 0;
                                break;

                            case ASN_OBJECTIDENTIFIER:  //0x06
                                i += asn_objectidentifier(Op,&buf[i],length);
                                length = 0;
                                break;

                            case ASN_ENUMERATED:        //0x0a
                                i += mheg_enumtype(Op,EventType,&buf[i],length);
                                length = 0;
                                break;

                            default:
                                printf("\n *** undefined universal primitive, %02x ***\n",tag);
                                i ++;
                                length --;
                                break;
                        }
//fprintf(Op,"[(1)enable_actionslots=%d]",enable_actionslots);
                        break;

                    case ASN_CONSTRUCTOR:
//fprintf(Op,"[(8)enable_actionslots=%d]",enable_actionslots);
                        switch (encoding | tag)
                        {
                            case ASN_SEQUENCE:          //0x30
                                if (asn_debug & 1) printf("[%02x] ",buf[i]);
//                                printf("ASN_SEQUENCE    ");
                                j=0;

                                // hack to add :ActionSlots ...
                                if (enable_actionslots == 1)
                                {
//fprintf(Op,"[enable_actionslots==1,len=%d]",length);
//                                    enable_actionslots = 2;
                                    fprintf(Op,"\n%s( ",tab());

//                                    fprintf(Op,"\n%s:ActionSlots ",tab());
                                    enable_actionslots++;
                                    tabl ++;
                                    if (j < length)
                                        j += asn_decode(Op,&buf[i+j]);
                                    fprintf(Op,"\n%s:ActionSlots ",tab());
                                    tabl --;
                                    enable_actionslots--;
//                                    fprintf(Op,"\n%s) ",tab());

//                                    enable_actionslots = 2;
                                }
/*
                                else if (enable_actionslots == 2)
                                {
fprintf(Op,"[enable_actionslots==2,len=%d]",length);
//                                    fprintf(Op,"\n%s:ActionSlots ( ",tab());
//                                    enable_actionslots = 0;
                                    enable_actionslots ++;
                                    fprintf(Op,"\n%s( ",tab());
                                }
                                else if (enable_actionslots == 3)
                                {
//fprintf(Op,"[enable_actionslots==3,len=%d]",length);
                                    fprintf(Op,"\n%s( ",tab());
//                                    fprintf(Op,"\n%s:ActionSlots ( ",tab());
                                    enable_actionslots = 0;
                                }
*/
                                // need a :ContentRef after a OrigContent
                                else if (last_tag == 0x3a)
                                {
                                     fprintf(Op,"\n%s:ContentRef ( ",tab());
                                }
                                else
                                {
                                    fprintf(Op,"\n%s( ",tab());
                                }

                                enable_actionslots++;
                                tabl ++;
//fprintf(Op,"[length=%x]",length);
//                                j=0;
                                while (j < length)
                                {
//if (enable_actionslots) enable_actionslots++;
//fprintf(Op,"[(6)enable_actionslots=%d]",enable_actionslots);
                                    j += asn_decode(Op,&buf[i+j]);
                                }
                                tabl --;
                                enable_actionslots--;
//                                fprintf(Op,") ");
                                fprintf(Op,"\n%s) ",tab());
//fprintf(Op,"[(2)enable_actionslots=%d]",enable_actionslots);
//if (enable_actionslots == 1) enable_actionslots = 0;

                                i +=length;
                                length = 0;
                                break;
                            default:
                                printf("\n *** undefined universal constructor, %02x ***\n",tag);
                                i ++;
                                length --;
                                break;
//printf("universal constructor, 0x%02x (%d) ***\n",tag,tag);
                        }
//fprintf(Op,"[(2)enable_actionslots=%d]",enable_actionslots);
                }
                break;

            case 0x40:  // application tag
                printf("\n *** undefined application tag, %02x ***\n",tag);
                i ++;
                length --;
                break;

            case 0x80:  // private tag
                last_tag = tag;
                switch(encoding)
                {
                        case ASN_PRIMITIVE:

                            fprintf(Op,"\n%s:",tab());
                            tabl ++;
                            // hack for ContentRef
                            if (tag == 0x45) // ContentRef
                            {
                                fprintf(Op,"%s ",getClass(Op,0xea));    //NewRefContent
                            }
                            else
                            {
                                fprintf(Op,"%s ",getClass(Op,tag));
                            }

                            switch(tag)
                            {
                                // boolean
                                case 0x03: // StandardVersion
                                case 0x07: // OrigGCPriority
                                case 0x36:
                                case 0x38:
                                case 0x3b:
                                case 0x3d: // CCPriority
                                case 0x41: // InitiallyAvailable
                                case 0x4a:
                                case 0x4b:
                                case 0x4f:
                                case 0x51:
                                case 0x5b:
                                case 0x65:
                                case 0x6f:
                                case 0x91:
                                    if (length == 1) i += asn_boolean(Op,&buf[i]);
                                    else i += asn_integer(Op,&buf[i],length);
                                    length = 0;
                                    break;

                                // integer
                                case 0x26:
                                case 0x28:
                                case 0x2c:
                                case 0x2d:
                                case 0x2e:  // BitmapCHook
                                case 0x2f:
                                case 0x33:
                                case 0x39:  // CHook
                                case 0x50:
                                case 0x52:
                                case 0x5e:  // Looping
                                case 0x62:  // ComponentTag
                                case 0x63:  // OriginalVolume
                                case 0x67:
                                case 0x68:
                                case 0x69:
                                case 0x6a:
                                case 0x6b:
                                case 0x70:
//                                    if (length == 1)                           // not for chook,
//                                        i += asn_boolean(Op,&buf[i]);
//                                    else
                                        i += asn_integer(Op,&buf[i],length);
                                    length = 0;
                                    break;


                                // octet string
                                case 0x04: // ObjectInfo
                                case 0x40:
                                case 0x45: // ContentRef
                                case 0x6e:
                                case 0x71:
                                    i += asn_octetstring(Op,&buf[i],length);
                                    length = 0;
                                    break;
                                case 0x2b: // FontAttributes
                                    if (maja)
                                    {
                                        i += asn_octetstring_adjust(Op,&buf[i],length);
                                    }
                                    else
                                    {
                                    i += asn_octetstring(Op,&buf[i],length);
                                    }
                                    font_attribute_found = 1;
                                    length = 0;
                                    break;


                                // Justification
                                case 0x57:
                                case 0x58:
                                    i += mheg_enumtype(Op,Justification,&buf[i],length);
                                    length = 0;
                                    break;

                                // LineOrientation
                                case 0x59:
                                    i += mheg_enumtype(Op,LineOrientation,&buf[i],length);
                                    length = 0;
                                    break;

                                // StartCorner
                                case 0x5a:
                                    i += mheg_enumtype(Op,StartCorner,&buf[i],length);
                                    length = 0;
                                    break;

                                // Storage
                                case 0x5d:
                                    i += mheg_enumtype(Op,Storage,&buf[i],length);
                                    length = 0;
                                    break;

                                // Termination
                                case 0x64:
                                    i += mheg_enumtype(Op,Termination,&buf[i],length);
                                    length = 0;
                                    break;

                                // Orientation
                                case 0x66:
                                    i += mheg_enumtype(Op,Orientation,&buf[i],length);
                                    length = 0;
                                    break;

                                // SliderStyle
                                case 0x6c:
                                    i += mheg_enumtype(Op,SliderStyle,&buf[i],length);
                                    length = 0;
                                    break;

                                // InputType
                                case 0x6d:
                                    i += mheg_enumtype(Op,InputType,&buf[i],length);
                                    length = 0;
                                    break;

                                // ButtonStyle
                                case 0x72:
                                    i += mheg_enumtype(Op,ButtonStyle,&buf[i],length);
                                    length = 0;
                                    break;

                                default:
                                    printf(" *** undefined private primitive, 0x%02x (%d)***\n",tag,tag);
                                    length = 0;
                                    break;
                                // etc
                            }
//fprintf(Op,"[(3)enable_actionslots=%d]",enable_actionslots);
                            tabl --;
//printf("private primitive, 0x%02x (%d) ***\n",tag,tag);
                            break;

                        case ASN_CONSTRUCTOR:
                            switch(tag)
                            {
                                // opens a curly bracket ...
                                case 0x00:    // application
                                    get_default_attributes(&buf[i],length);
                                case 0x01:    // scene
                                case 0x09:
                                case 0x0e:
                                case 0x0f:
                                case 0x10:
                                case 0x11:
                                case 0x12:
                                case 0x13:
                                case 0x14:
                                case 0x15:
                                case 0x16:
                                case 0x18:    // DynamicLineArt
                                case 0x19:
                                case 0x1a:    // Hotspot
                                case 0x1e:
                                case 0x20:    // Slider
                                case 0x21:
                                case 0x22:    // ListGroup
                                case 0x5f:    // Audio
                                case 0x60:    // Video
                                    fprintf(Op,"\n%s{:",tab());
                                    tabl ++;
                                    fprintf(Op,"%s ",getClass(Op,tag));

                                    j=0;
                                    while (j < length)
                                    {
                                        j += asn_decode(Op,&buf[i+j]);
                                    }
                                    tabl --;
                                    fprintf(Op,"\n%s}",tab());
                                    i +=length;
                                    length = 0;
                                    break;

                                case 0x1d:    // Text
                                    fprintf(Op,"\n%s{:",tab());
                                    tabl ++;
                                    fprintf(Op,"%s ",getClass(Op,tag));

                                    font_attribute_found = 0;
                                    text_colour_found = 0;

                                    j=0;
                                    while (j < length)
                                    {
                                        j += asn_decode(Op,&buf[i+j]);
                                    }

                                    // set default attributes...
                                    // but only if not already set
                                    if (maja && default_font_attribute_length && !font_attribute_found)
                                    {
                                        fprintf(Op,"\n%s:FontAttributes ",tab());
                                        asn_octetstring_adjust(Op,default_font_attribute,default_font_attribute_length);
                                    }
                                    if (maja && default_text_colour_length && !text_colour_found)
                                    {
                                        fprintf(Op,"\n%s:TextColour ",tab());
                                        asn_decode(Op, default_text_colour);
                                    }

                                    tabl --;
                                    fprintf(Op,"\n%s}",tab());
                                    i +=length;
                                    length = 0;
                                    break;



                                // opens a bracket ...
                                case 0x08: // Items
                                case 0x3f: // LinkEffect
                                case 0x46: // MovementTable
//                                case 0x47: // TokenGroupItems
                                case 0x48: // ActionSlots
                                case 0x49: // Positions
                                case 0x5c: // Multiplex
                                    fprintf(Op,"\n%s:",tab());
                                    tabl ++;
                                    fprintf(Op,"%s ",getClass(Op,tag));
                                    fprintf(Op,"( ");

                                    j=0;
                                    while (j < length)
                                    {
//                                        printf("\n%s",tab());
                                        j += asn_decode(Op,&buf[i+j]);
                                    }
                                    tabl --;
                                    fprintf(Op,"\n%s)",tab());
                                    i +=length;
                                    length = 0;
                                    break;



                                // opens bracket, same line
                                case 0x37: // NextScenes
                                case 0x73: // Activate
                                case 0x74: // Add
                                case 0x75: // AddItem
                                case 0x76: // Append
                                case 0x77: // BringToFront
                                case 0x78: // Call
                                case 0x79: // CallActionSlot
                                case 0x7d: // Deactivate
                                case 0x81: // Divide
                                case 0x85: // DrawPolygon
                                case 0x86: // DrawPolyLine
                                case 0x89: // Fork
                                case 0x8c: // GetCellItem
                                case 0x8e: // GetEngineSupport
                                case 0x91: // GetFirstitem
                                case 0x9a: // GetLineItem
                                case 0x9e: // GetPosition
                                case 0x9f: // GetRunningStatus
                                case 0xa1: // GetSlidervalue
                                case 0xa3: // GettextData
                                case 0xa4: // GetTokenPosition
                                case 0xa6: // Launch
                                case 0xa8: // Modulo
                                case 0xa9: // Move
                                case 0xaa: // MoveTo
                                case 0xab: // Multiply
                                case 0xad: // Preload
                                case 0xae: // PutBefore
                                case 0xaf: // PutBehind
                                case 0xb0: // Quit
                                case 0xb1: // ReadPersistent
                                case 0xb2: // Run
                                case 0xb3: // ScaleBitmap
                                case 0xb4: // ScaleVideo
                                case 0xb5: // ScrollItems
                                case 0xb6: // Select
                                case 0xb8: // SendEvent
                                case 0xb9: // SendToBack
                                case 0xba: // SetBoxSize
                                case 0xbb: // SetCachePriority
                                case 0xc1: // SetData
                                case 0xc2: // SetEntryPoint
                                case 0xc3: // SetFillColour
                                case 0xc6: // SetHighlightStatus
                                case 0xc7: // SetInteractionStatus
                                case 0xc9: // SetLineColour
                                case 0xcb: // SetLineWidth
                                case 0xcc: // SetOverwriteMode
                                case 0xce: // SetPortion
                                case 0xcf: // SetPosition
                                case 0xd2: // SetTimer
                                case 0xd4: // SetVariable
                                case 0xd6: // Spawn
                                case 0xd7: // Step
                                case 0xd8: // Stop
                                case 0xd9: // StorePersistent
                                case 0xda: // Subtract
                                case 0xdb: // TestVariable
                                case 0xde: // TransitionTo
                                case 0xdf: // Unload
                                case 0xe0: // UnlockScreen
                                    fprintf(Op,"\n%s:",tab());
                                    fprintf(Op,"%s ",getClass(Op,tag));
                                    fprintf(Op,"( ");

                                    j=0;
                                    while (j < length)
                                    {
                                        j += asn_decode(Op,&buf[i+j]);
                                    }
                                    fprintf(Op,") ");
                                    i +=length;
                                    length = 0;
                                    break;

                                // plain tag, :, new line
                                case 0x29: // TextColour
                                    text_colour_found = 1;
                                case 0x02:    // StdID
//                                case 0x25:
                                case 0x27:
                                case 0x2a: // Font
                                case 0x31: // HighlightRefColour
                                case 0x32: // SliderRefColour
                                case 0x34: // SceneCS
                                case 0x3a:
                                case 0x43:
                                case 0x44:
                                case 0x4c: // OrigBoxSize
                                case 0x4d: // OrigPosition
                                case 0x54:
                                case 0x55:
                                case 0xa7:
                                    fprintf(Op,"\n%s:",tab());
                                    tabl ++;
                                    fprintf(Op,"%s ",getClass(Op,tag));
                                    j=0;
                                    while (j < length)
                                    {
                                        j += asn_decode(Op,&buf[i+j]);
                                    }
                                    tabl --;
                                    i +=length;
                                    length = 0;
                                    break;

                                // no initial new line
                                case 0xe1: // GBoolean
                                case 0xe2: // GInteger
                                case 0xe3:
                                case 0xe4:
                                case 0xe5:
                                case 0xe7:
                                case 0xeb: // NewCCPriority
                                case 0xec:
                                    fprintf(Op,":");
                                    fprintf(Op,"%s ",getClass(Op,tag));
                                    j=0;
                                    while (j < length)
                                    {
                                        j += asn_decode(Op,&buf[i+j]);
                                    }
                                    i +=length;
                                    length = 0;
                                    break;

                                // LinkCondition
                                case 0x3e: // LinkCondition
                                    fprintf(Op,"\n%s:",tab());
                                    fprintf(Op,"EventSource ");
                                    j = 0;
                                    j = asn_decode(Op,&buf[i+j]);
                                    fprintf(Op,"\n%s:",tab());
                                    fprintf(Op,"EventType ");
                                    j += asn_decode(Op,&buf[i+j]);
                                    while (j < length)
                                    {
                                        fprintf(Op,"\n%s:",tab());
                                        fprintf(Op,"EventData ");
                                        j += asn_decode(Op,&buf[i+j]);
                                    }
                                    i +=length;
                                    length = 0;
                                    break;

                                // TokenGroupItems
                                case 0x47: // TokenGroupItems
// :TokenGroupItems ( ( 6101 :ActionSlots ( ( :Call ... ) ) )  ( 6101 :Action...) )
//                                asn_debug = 1;
//fprintf(Op,"[ActionSlots1]");
                                    fprintf(Op,"\n%s:",tab());
                                    tabl ++;
                                    fprintf(Op,"%s ",getClass(Op,tag));
                                    fprintf(Op,"( ");


                                    j=0;
                                    enable_actionslots = 1;
//                                    j += asn_decode(Op,&buf[i+j]);
// this needs to inside loop ...
//fprintf(Op,"[j=%x, length=%x]",j,length);
                                    while (j < length)
                                    {
//                                        enable_actionslots = 1;
                                        j += asn_decode(Op,&buf[i+j]);
//fprintf(Op,"[ActionSlots2]");
                                    }
                                    tabl --;
                                    fprintf(Op,"\n%s)",tab());
                                    i +=length;
                                    enable_actionslots = 0;
                                    length = 0;
//                                asn_debug = 0;
                                    break;


                                // hack for NewRefContent - ignore
                                case 0xea: // NewRefContent
                                    i +=length;
                                    length = 0;
                                    break;


                                // do nothing ?
                                case 0x25: // DefaultAttributes
                                    length = 0;
                                    break;

                                // etc
                                default:
                                    printf(" *** undefined private constructor, 0x%02x (%d) ***\n",tag,tag);
                                    fprintf(Op,"\n[ *** undefined private constructor, 0x%02x (%d) ***] ",tag,tag);
                                    length = 0;
                                    break;
                            }
//printf("private constructor, 0x%02x (%d) ***\n",tag,tag);
//fprintf(Op,"[(4)enable_actionslots=%d]",enable_actionslots);
                            break;
                }
                break;

            case 0xc0:  // context tag
                printf("\n *** undefined context tag, %02x ***\n",tag);
                i ++;
                length --;
                break;
        }
    }
    while(length);

//if (enable_actionslots) enable_actionslots++;

//if (enable_actionslots>=2)
//{
//fprintf(Op,"[(5)enable_actionslots=%d]",enable_actionslots);
//    fprintf(Op,":ActionSlots");
//    enable_actionslots=0;
//}


        // end of content
        // printf("[%02x] ",buf[i]);


//printf("\n%s ",tab());
//printf("return ",);
//    tabl --;
    return i;
}

// quick scan to the default attributes section
// and pull out the default FontAttributes
int get_default_attributes(UCHAR *buf, int length1)
{
    int class, encoding, tag;
    int subsequent;
    int length;
    int i = 0;
    int j;
    int default_attributes = 0;

    default_font_attribute_length = 0;
    default_text_colour_length = 0;

//    printf("\nget_default_attributes(i=%d, len=%x (%d)) ",i,length1,length1);

    while (i<length1)
    {
        class = buf[i] & 0xc0;
        encoding = buf[i] & 0x20;
        tag = buf[i] & 0x1f;
        i++;

        // subsequent identifier octets
        if (tag == 0x1f)
        {
            tag = 0;
            do
            {
                tag = (tag << 7) + (buf[i] & 0x7f);
                subsequent = buf[i] & 0x80;
                i++;
            } while (subsequent);
        }


//        printf("\n[class:%02x, encoding:%02x, tag:%02x]",class,encoding,tag);


        // :DefaultAttributes
        if ((class==ASN_CONTEXT) && (encoding==ASN_CONSTRUCTOR) && (tag==0x25))
        {
//            printf("\n:DefaultAttributes ");
            default_attributes = 1;
            length = buf[i] & 0x7f;
//            printf("length:%02x",length);
            i++;
        }
        else
        {
            length = buf[i] & 0x7f;
            subsequent = buf[i] & 0x80;
            i++;

            // subsequent length
            if (subsequent)
            {
                subsequent = length;
                length = 0;
                do
                {
                    length = (length << 8) + (buf[i] & 0xff);
                    subsequent --;
                    i++;
                }
                while (subsequent);
            }
//            printf("length:%02x",length);

            // :FontAttributes
            if (default_attributes && (class==ASN_CONTEXT) && (encoding==ASN_PRIMITIVE) && (tag==0x2b))
            {
                // octet_string
                default_font_attribute_length=length;
                for (j=0; j<length; j++)
                {
                    default_font_attribute[j] = buf[i+j];
//                    printf("[%02x]",buf[i+j]);
                }
            }
            // :TextColour
            if (default_attributes && (class==ASN_CONTEXT) && (encoding==ASN_CONSTRUCTOR) && (tag==0x29))
            {
                default_text_colour_length=length;
                for (j=0; j<length; j++)
                {
                    default_text_colour[j] = buf[i+j];
//                    printf("[%02x]",buf[i+j]);
                }
            }

            i += length;
        }
    }


    return 0;
}

